// decodable-nnet-online-for-hark.cc
// If you want to put this file under the directory, it file is placed in under the "nnet" directory.

#ifdef _MSC_VER
#define NOMINMAX
#endif
#include "decodable-nnet2-online-for-hark.h"

//#include "nnet2/online-nnet2-decodable.h"
#include "nnet3/nnet-utils.h"

namespace kaldi {
namespace nnet2 {

DecodableNnet2OnlineForHark::DecodableNnet2OnlineForHark(
        const kaldi::nnet2::AmNnet &nnet,
        const kaldi::TransitionModel &trans_model,
        const DecodableNnet2OnlineForHarkOptions &opts,
        //kaldi::OnlineFeatureInterface *features_if):
        std::shared_ptr<kaldi::OnlineFeatureInterface> features_if):
    features_if_(features_if),
    nnet_(nnet),
    trans_model_(trans_model),
    opts_(opts),
    feat_dim_(features_if->Dim()),
    left_context_(nnet.GetNnet().LeftContext()),
    right_context_(nnet.GetNnet().RightContext()),
    num_pdfs_(nnet.GetNnet().OutputDim()),
    begin_frame_(-1) {
  KALDI_ASSERT(opts_.max_nnet_batch_size > 0);
  log_priors_ = nnet_.Priors();
  KALDI_ASSERT(log_priors_.Dim() == trans_model_.NumPdfs() &&
               "Priors in neural network not set up (or mismatch with transition model).");
  log_priors_.ApplyLog();
}

kaldi::BaseFloat DecodableNnet2OnlineForHark::LogLikelihood(kaldi::int32 frame, kaldi::int32 index) {
  ComputeForFrame(frame);
  kaldi::int32 pdf_id = trans_model_.TransitionIdToPdf(index);
  KALDI_ASSERT(frame >= begin_frame_ && (frame < begin_frame_ + scaled_loglikes_.NumRows()));
  return scaled_loglikes_(frame - begin_frame_, pdf_id);
}

bool DecodableNnet2OnlineForHark::IsLastFrame(kaldi::int32 frame) const {
  //opts_.pad_input == true is normal case
  return features_if_->IsLastFrame(frame + (opts_.pad_input ? 0 : left_context_ + right_context_));
}

kaldi::int32 DecodableNnet2OnlineForHark::NumFramesReady() const {
  kaldi::int32 features_ready = features_if_->NumFramesReady();
  if(features_ready == 0){ return 0; }
  bool input_finished = features_if_->IsLastFrame(features_ready - 1);
  if(opts_.pad_input){
    // normal case... we'll pad with duplicates of first + last frame to get the
    // required left and right context.
    if(input_finished){ return features_ready; }
    else{ return std::max<kaldi::int32>(0, features_ready - right_context_); }
  }
  else{
    return std::max<kaldi::int32>(0, features_ready - right_context_ - left_context_);
  }
}

void DecodableNnet2OnlineForHark::ComputeForFrame(kaldi::int32 frame) {
  kaldi::int32 features_ready = features_if_->NumFramesReady();
  bool input_finished = features_if_->IsLastFrame(features_ready - 1);  
  KALDI_ASSERT(frame >= 0);
  if(frame >= begin_frame_ && (frame < begin_frame_ + scaled_loglikes_.NumRows())){
    return;
  }
  KALDI_ASSERT(frame < NumFramesReady());

  kaldi::int32 input_frame_begin = frame - (opts_.pad_input ? left_context_ : 0);
  kaldi::int32 max_possible_input_frame_end = features_ready + ((input_finished && opts_.pad_input) ? right_context_ : 0);
  kaldi::int32 input_frame_end =
      std::min<kaldi::int32>(max_possible_input_frame_end,
                             input_frame_begin + left_context_ + right_context_ + opts_.max_nnet_batch_size);

  KALDI_ASSERT(input_frame_end > input_frame_begin);
  kaldi::Matrix<kaldi::BaseFloat> features(input_frame_end - input_frame_begin, feat_dim_);
  for(kaldi::int32 t = input_frame_begin; t < input_frame_end; t++){
    kaldi::SubVector<kaldi::BaseFloat> row(features, t - input_frame_begin);
    kaldi::int32 t_modified = (t < 0 ? 0 : (t >= features_ready ? features_ready - 1 : t)); // This statements take care of "pad_input"
    features_if_->GetFrame(t_modified, &row);
  }
  kaldi::CuMatrix<kaldi::BaseFloat> cu_features; 
  cu_features.Swap(&features);  // Copy to GPU, if we're using one.

  kaldi::int32 num_frames_out = input_frame_end - input_frame_begin - left_context_ - right_context_;
  kaldi::CuMatrix<kaldi::BaseFloat> cu_posteriors(num_frames_out, num_pdfs_);

  // The "false" below tells it not to pad the input: we've already done
  // any padding that we needed to do.
  kaldi::nnet2::NnetComputation(nnet_.GetNnet(), cu_features, false, &cu_posteriors);
  
  cu_posteriors.ApplyFloor(1.0e-20); // Avoid log of zero which leads to NaN.
  cu_posteriors.ApplyLog();
  // subtract log-prior (divide by prior)
  cu_posteriors.AddVecToRows(-1.0, log_priors_);
  // apply probability scale.
  cu_posteriors.Scale(opts_.acoustic_scale);

  // Transfer the scores the CPU for faster access by the decoding process.
  scaled_loglikes_.Resize(0, 0);
  cu_posteriors.Swap(&scaled_loglikes_);

  begin_frame_ = frame;
}


} // namespace nnet2
} // namespace kaldi
