
#ifdef _MSC_VER
#define NOMINMAX
#endif

#include "decodable-nnet3-online-for-hark.h"

namespace kaldi {
namespace nnet3 {

DecodableNnet3SimpleOnlineForHark::DecodableNnet3SimpleOnlineForHark(
        const DecodableNnetSimpleLoopedInfo &decodable_info,
        const kaldi::TransitionModel &trans_model,
        const NnetSimpleLoopedComputationOptions &opts,
        kaldi::OnlineFeatureInterface *features_if,
        kaldi::OnlineFeatureInterface *ivector_if):
    num_chunks_computed_(0),
    current_log_post_subsampled_offset_(-1),
    decodable_info_(decodable_info),
    trans_model_(trans_model),
    opts_(opts),
    features_if_(features_if),
    ivector_if_(ivector_if),
    computer_(decodable_info_.opts.compute_config, decodable_info_.computation, decodable_info_.nnet, NULL) {
  KALDI_ASSERT(features_if_ != nullptr);

  kaldi::int32 nnet_feature_dim = decodable_info_.nnet.InputDim("input");
  kaldi::int32 nnet_ivector_dim = decodable_info_.nnet.InputDim("ivector");
  kaldi::int32 feature_dim = features_if_->Dim();
  kaldi::int32 ivector_dim = (ivector_if_ != nullptr) ? ivector_if_->Dim() : -1;
  if(nnet_feature_dim != feature_dim){
    KALDI_ERR << "Input feature dimension mismatch: got " << feature_dim
              << " but network expects " << nnet_feature_dim;
  }
  if(nnet_ivector_dim != ivector_dim){
    KALDI_ERR << "Ivector feature dimension mismatch: got " << ivector_dim
              << " but network expects " << nnet_ivector_dim;
  }
}

kaldi::int32 DecodableNnet3SimpleOnlineForHark::NumFramesReady() const {
  // note: the ivector_if_ may have 2 or 3 fewer frames ready than
  // input_features_, but we don't wait for them; we just use the most recent
  // iVector we can.
  kaldi::int32 features_ready = features_if_->NumFramesReady();
  if(features_ready == 0){ return 0; }
  bool input_finished = features_if_->IsLastFrame(features_ready - 1);

  kaldi::int32 sf = decodable_info_.opts.frame_subsampling_factor;

  if(input_finished){
    // if the input has finished,... we'll pad with duplicates of the last frame
    // as needed to get the required right context.
    return (features_ready + sf - 1) / sf;
  }
  else{
    // note: info_.right_context_ includes both the model context and any
    // extra_right_context_ (but this
    kaldi::int32 non_subsampled_output_frames_ready = std::max<int32>(0, features_ready - decodable_info_.frames_right_context);
    kaldi::int32 num_chunks_ready = non_subsampled_output_frames_ready / decodable_info_.frames_per_chunk;

    // note: the division by the frame subsampling factor 'sf' below
    // doesn't need any attention to rounding because info_.frames_per_chunk
    // is always a multiple of 'sf' (see 'frames_per_chunk = GetChunksize..."
    // in decodable-simple-looped.cc).
    return num_chunks_ready * decodable_info_.frames_per_chunk / sf;
  }
}

// note: the frame-index argument is on the output of the network, i.e. after any
// subsampling, so we call it 'subsampled_frame'.
bool DecodableNnet3SimpleOnlineForHark::IsLastFrame(kaldi::int32 subsampled_frame) const {
  // To understand this code, compare it with the code of NumFramesReady(),
  // it follows the same structure.
  kaldi::int32 features_ready = features_if_->NumFramesReady();
  if(features_ready == 0){
    if(subsampled_frame == -1 && features_if_->IsLastFrame(-1)){
      // the attempt to handle this rather pathological case (input finished
      // but no frames ready) is a little quixotic as we have not properly
      // tested this and other parts of the code may die.
      return true;
    }
    else{
      return false;
    }
  }
  bool input_finished = features_if_->IsLastFrame(features_ready - 1);
  if(!input_finished){ return false; }
  kaldi::int32 sf = decodable_info_.opts.frame_subsampling_factor;
  kaldi::int32 num_subsampled_frames_ready = (features_ready + sf - 1) / sf;
  return (subsampled_frame == num_subsampled_frames_ready - 1);
}

void DecodableNnet3SimpleOnlineForHark::AdvanceChunk() {
  // Prepare the input data for the next chunk of features.
  // note: 'end' means one past the last.
  kaldi::int32 begin_input_frame, end_input_frame;
  if(num_chunks_computed_ == 0){
    begin_input_frame = -decodable_info_.frames_left_context;
    // note: end is last plus one.
    end_input_frame = decodable_info_.frames_per_chunk + decodable_info_.frames_right_context;
  }
  else{
    // note: begin_input_frame will be the same as the previous end_input_frame.
    // you can verify this directly if num_chunks_computed_ == 0, and then by
    // induction.
    begin_input_frame = num_chunks_computed_ * decodable_info_.frames_per_chunk + decodable_info_.frames_right_context;
    end_input_frame = begin_input_frame + decodable_info_.frames_per_chunk;
  }

  kaldi::int32 num_feature_frames_ready = features_if_->NumFramesReady();
  bool is_finished = features_if_->IsLastFrame(num_feature_frames_ready - 1);

  if(end_input_frame > num_feature_frames_ready && !is_finished){
    // we shouldn't be attempting to read past the end of the available features
    // until we have reached the end of the input (i.e. the end-user called
    // InputFinished(), announcing that there is no more waveform; at this point
    // we pad as needed with copies of the last frame, to flush out the last of
    // the output.
    // If the following error happens, it likely indicates a bug in this
    // decodable code somewhere (although it could possibly indicate the
    // user asking for a frame that was not ready, which would be a misuse
    // of this class.. it can be figured out from gdb as in either case it
    // would be a bug in the code.
    KALDI_ERR << "Attempt to access frame past the end of the available input";
  }

  kaldi::CuMatrix<kaldi::BaseFloat> feats_chunk;
  { // this block sets 'feats_chunk'.
    kaldi::Matrix<kaldi::BaseFloat> this_feats(end_input_frame - begin_input_frame, features_if_->Dim());
    for(kaldi::int32 i = begin_input_frame; i < end_input_frame; i++){
      kaldi::SubVector<kaldi::BaseFloat> this_row(this_feats, i - begin_input_frame);
      kaldi::int32 input_frame = i;
      if(input_frame < 0){ input_frame = 0; }
      if(input_frame >= num_feature_frames_ready){ input_frame = num_feature_frames_ready - 1; }
      features_if_->GetFrame(input_frame, &this_row);
    }
    feats_chunk.Swap(&this_feats);
  }
  computer_.AcceptInput("input", &feats_chunk);

  if(decodable_info_.has_ivectors){
    KALDI_ASSERT(ivector_if_ != NULL);
    KALDI_ASSERT(decodable_info_.request1.inputs.size() == 2);
    // all but the 1st chunk should have 1 iVector, but there is no need to
    // assume this.
    kaldi::int32 num_ivectors = (num_chunks_computed_ == 0 ? decodable_info_.request1.inputs[1].indexes.size() : decodable_info_.request2.inputs[1].indexes.size());
    KALDI_ASSERT(num_ivectors > 0);

    kaldi::Vector<kaldi::BaseFloat> ivector(ivector_if_->Dim());
    // we just get the iVector from the last input frame we needed,
    // reduced as necessary
    // we don't bother trying to be 'accurate' in getting the iVectors
    // for their 'correct' frames, because in general using the
    // iVector from as large 't' as possible will be better.

    kaldi::int32 most_recent_input_frame = num_feature_frames_ready - 1,
      num_ivector_frames_ready = ivector_if_->NumFramesReady();

    if(num_ivector_frames_ready > 0){
      kaldi::int32 ivector_frame_to_use = std::min<int32>(most_recent_input_frame, num_ivector_frames_ready - 1);
      ivector_if_->GetFrame(ivector_frame_to_use, &ivector);
    }
    // else just leave the iVector zero (would only happen with very small
    // chunk-size, like a chunk size of 2 which would be very inefficient; and
    // only at file begin.

    // note: we expect num_ivectors to be 1 in practice.
    kaldi::Matrix<kaldi::BaseFloat> ivectors(num_ivectors, ivector.Dim());
    ivectors.CopyRowsFromVec(ivector);
    kaldi::CuMatrix<kaldi::BaseFloat> cu_ivectors;
    cu_ivectors.Swap(&ivectors);
    computer_.AcceptInput("ivector", &cu_ivectors);
  }
  computer_.Run();

  {
    // Note: it's possible in theory that if you had weird recurrence that went
    // directly from the output, the call to GetOutputDestructive() would cause
    // a crash on the next chunk.  If that happens, GetOutput() should be used
    // instead of GetOutputDestructive().  But we don't anticipate this will
    // happen in practice.
    kaldi::CuMatrix<kaldi::BaseFloat> output;
    computer_.GetOutputDestructive("output", &output);

    if(decodable_info_.log_priors.Dim() != 0){
      // subtract log-prior (divide by prior)
      output.AddVecToRows(-1.0, decodable_info_.log_priors);
    }
    // apply the acoustic scale
    output.Scale(decodable_info_.opts.acoustic_scale);
    current_log_post_.Resize(0, 0);
    current_log_post_.Swap(&output);
  }
  KALDI_ASSERT(current_log_post_.NumRows() == decodable_info_.frames_per_chunk / decodable_info_.opts.frame_subsampling_factor &&
               current_log_post_.NumCols() == decodable_info_.output_dim);

  num_chunks_computed_++;

  current_log_post_subsampled_offset_ = (num_chunks_computed_ - 1) * (decodable_info_.frames_per_chunk / decodable_info_.opts.frame_subsampling_factor);
}

kaldi::BaseFloat DecodableNnet3SimpleOnlineForHark::LogLikelihood(kaldi::int32 subsampled_frame, kaldi::int32 index){
  EnsureFrameIsComputed(subsampled_frame);
  return current_log_post_(subsampled_frame - current_log_post_subsampled_offset_, trans_model_.TransitionIdToPdf(index));
}

} // namespace nnet3
} // namespace kaldi
