
#ifndef MP_TIME_WRAPPER_H_
#define MP_TIME_WRAPPER_H_

#include <ctime>
#include <iomanip>
#include <sstream>
#include <string>

#ifdef _MSC_VER
#define WIN32_LEAN_AND_MEAN
#define NOMINMAX
#include <windows.h>
#else // MSCVER
#include <sys/time.h>
#endif // _MSC_VER


namespace multiplatform {
namespace time {

#ifdef _MSC_VER

#ifndef _TIMEVAL_DEFINED
#define _TIMEVAL_DEFINED
struct timeval {
  time_t tv_sec;
  long   tv_usec;
};
#endif // _TIMEVAL_DEFINED

#ifndef _TIMEZONE_DEFINED
#define _TIMEZONE_DEFINED
struct timezone {
  int tz_minuteswest;
  int tz_dsttime;
};
#endif // _TIMEZONE_DEFINED

#ifndef HAVE_STRUCT_TIMESPEC
#define HAVE_STRUCT_TIMESPEC
#ifndef _TIMESPEC_DEFINED
#define _TIMESPEC_DEFINED
struct timespec {
  time_t tv_sec;
  long   tv_nsec;
};
#endif // _TIMESPEC_DEFINED
#endif // HAVE_STRUCT_TIMESPEC

constexpr unsigned __int64 delta_epoch = 116444736000000000ULL;
//constexpr unsigned __int32 int32_exp6_expr = 1000000UL;
constexpr unsigned __int32 int32_exp7_expr = 10000000UL;
//constexpr unsigned __int32 int32_exp8_expr = 100000000UL;
constexpr unsigned __int32 int32_exp9_expr = 1000000000UL;

int gettimeofday(struct timeval *tv, struct timezone *tz){
  if(tv == nullptr){ return -1; }
  unsigned __int64 epoch = 0; // UNIX epoch (1970-01-01 00:00:00) expressed in Windows NT FILETIME
  FILETIME ft;
  GetSystemTimeAsFileTime(&ft);
  epoch |= ft.dwHighDateTime;
  epoch <<= 32;
  epoch |= ft.dwLowDateTime;
  epoch -= delta_epoch;
  //epoch /= 10;
  //tv->tv_sec = (epoch / int32_exp6_expr);
  //tv->tv_usec = (epoch % int32_exp6_expr);
  tv->tv_sec = (epoch / int32_exp7_expr);
  tv->tv_usec = (epoch % int32_exp7_expr) / 10;
  if(tz != nullptr){
    _tzset(); // !!CAUTION!! this function is thread unsafe.
    long l_timezone; _get_timezone(&l_timezone);
    int l_daylight; _get_daylight(&l_daylight);
    tz->tz_minuteswest = l_timezone / 60;
    tz->tz_dsttime = l_daylight;
  }
  return 0;
}

using clockid_t = int;
#ifdef ENABLE_HIGHRES_MODE
static bool g_time_initialized = false;
static LARGE_INTAGER g_cnt_per_sec;

int clock_gettime(clockid_t clk_id, struct timespec *tp){
  LARGE_INTEGER count;
  if(!g_time_initialized){
    g_time_initialized = true;
    if(QueryPerformanceFrequency(&g_cnt_per_sec) == 0){ g_cnt_per_sec.QuadPart = 0; }
  }
  if((tp == nullptr) || (g_cnt_per_sec.QuadPart <= 0) || (QueryPerformanceCounter(&count) == 0)){ return -1; }
  tp->tv_sec = count.QuadPart / g_cnt_per_sec.QuadPart;
  tp->tv_nsec = ((count.QuadPart % g_cnt_per_sec.QuadPart) * int32_exp9_expr) / g_cnt_per_sec.QuadPart;
}
#else // ENABLE_HIGHRES_MODE
int clock_gettime(clockid_t clk_id, struct timespec *tp){
  if(tp == nullptr){ return -1; }
  unsigned __int64 epoch = 0; // UNIX epoch (1970-01-01 00:00:00) expressed in Windows NT FILETIME
  FILETIME ft;
  GetSystemTimeAsFileTime(&ft);
  epoch |= ft.dwHighDateTime;
  epoch <<= 32;
  epoch |= ft.dwLowDateTime;
  epoch -= delta_epoch;
  tp->tv_sec = (epoch / int32_exp7_expr);
  tp->tv_nsec = (epoch % int32_exp7_expr) * 100;
  return 0;
}
#endif // ENABLE_HIGHRES_MODE

struct tm *localtime_r(const time_t *timep, struct tm *result){
  // Since you can pass output variables, you can use _localtime64_s() on Windows as well as localtime_r().
  _localtime64_s(result, timep);
  return result;
}

#endif // _MSC_VER

std::string formatted_time(){
  struct timeval tv;
  if(gettimeofday(&tv, nullptr)){ return "# CANNOT GET LOCAL TIME. #"; }
  std::tm broken;
  std::time_t time_t_sec = tv.tv_sec;
  localtime_r(&time_t_sec, &broken);
  if(!&broken){ return "# CANNOT GET LOCAL TIME. #"; }
  std::stringstream ss;
  ss << std::setw(4) << std::setfill(' ') << broken.tm_year + 1900 << "-"
     << std::setw(2) << std::setfill('0') << broken.tm_mon + 1 << "-"
     << std::setw(2) << std::setfill('0') << broken.tm_mday << " "
     << std::setw(2) << std::setfill('0') << broken.tm_hour << ":"
     << std::setw(2) << std::setfill('0') << broken.tm_min << ":"
     << std::setw(2) << std::setfill('0') << broken.tm_sec << "."
     << std::setw(6) << std::setfill('0') << tv.tv_usec;
  return ss.str();
}

} // namespace time
} // namespace multiplatform

#endif // MP_TIME_WRAPPER_H_

