// online-mfcnet.cc
// If you want to put this file under the directory, it file is placed in under the "feat" directory.

#ifdef _MSC_VER
#define NOMINMAX
#endif
#include "online-mfcnet.h"

namespace kaldi {

//FOR_FILE_MODE//

OnlineFeatureMfcnet::OnlineFeatureMfcnet(
    //kaldi::OnlineFeatureInputSettingOptions &opts,
    kaldi::OnlineNnet2FeaturePipelineInfo *feature_info,
    const kaldi::Matrix<kaldi::BaseFloat>& feats,
    std::uint32_t id = 0)
  : feature_info_(feature_info), socket_(nullptr), id_(id), feat_dim_(feats.NumCols()), feat_bytes_(0), num_frames_(feats.NumRows()), input_finished_(false)
{
  {
    source_info_.source_id = id;
    source_info_.azimuth = 0.0f;
    source_info_.elevation = 0.0f;
    source_info_.tv_sec = 0;
    source_info_.tv_usec = 0;
  }

  //DEBUG//{
  //DEBUG//  std::stringstream ss;
  //DEBUG//  ss << std::fixed << std::setprecision(6)
  //DEBUG//     << "source_id = "   << source_info_.source_id
  //DEBUG//     << ", azimuth = "   << source_info_.azimuth
  //DEBUG//     << ", elevation = " << source_info_.elevation
  //DEBUG//     << ", sec = "       << source_info_.tv_sec
  //DEBUG//     << ", usec = "      << source_info_.tv_usec;
  //DEBUG//  _raw_print_with_endl(ss.str());
  //DEBUG//}

  {
    features_ = feats;
    input_finished_ = true;
    //KALDI_LOG << "dim = " << feat_dim_;
  }

  if(feature_info_->feature_type.empty()){
    frame_shift_in_seconds_ = 0.01f;
  }
  else if(feature_info_->feature_type == "mfcc"){
    frame_shift_in_seconds_ = feature_info_->mfcc_opts.frame_opts.frame_shift_ms / 1000.0f;
  }
  else if(feature_info_->feature_type == "plp"){
    frame_shift_in_seconds_ = feature_info_->plp_opts.frame_opts.frame_shift_ms / 1000.0f;
  }
  else if(feature_info_->feature_type == "fbank"){
    frame_shift_in_seconds_ = feature_info_->fbank_opts.frame_opts.frame_shift_ms / 1000.0f;
  }
  else{
    frame_shift_in_seconds_ = 0.01f; // unknown feature type uses default value.
 }

}


OnlineFeatureMfcnet::OnlineFeatureMfcnet(
    //kaldi::OnlineFeatureInputSettingOptions &opts,
    kaldi::OnlineNnet2FeaturePipelineInfo *feature_info,
    std::shared_ptr<mfcnet_server> socket,
    std::uint32_t id = 0)
  : feature_info_(feature_info), socket_(socket), id_(id), feat_dim_(0), feat_bytes_(0), num_frames_(0), input_finished_(false)
{
  if(socket_ == nullptr){ return; }

  do{
    source_info_ = socket_->buffers[id_]->get_source();
    if(socket_->buffers[id_]->srcinfo_is_ready()){ break; } // The SorceInfo has been prepared, so it exit.
    multiplatform::time::ms_sleep(10); // It wait only when a retry occurs.
    _warning_print("srcinfo is not ready");
  } while(true);

  //DEBUG//{
  //DEBUG//  std::stringstream ss;
  //DEBUG//  ss << std::fixed << std::setprecision(6)
  //DEBUG//     << "source_id = "   << source_info_.source_id
  //DEBUG//     << ", azimuth = "   << source_info_.azimuth
  //DEBUG//     << ", elevation = " << source_info_.elevation
  //DEBUG//     << ", sec = "       << source_info_.tv_sec
  //DEBUG//     << ", usec = "      << source_info_.tv_usec;
  //DEBUG//  _raw_print_with_endl(ss.str());
  //DEBUG//}

  do{
    // The purpose is achieved by only updating the number of frames that can be acquired,
    // so the features are discarded.
    std::vector<float> features = socket_->buffers[id_]->get_feature();

    // Normal case:
    // The available features exist, so it exit.
    if(socket_->buffers[id_]->srcdata_is_ready()){ break; }

    // Error case:
    // Even if an available feature does not exist, reception processing can not be continued, so it exit.
    if(socket_->buffers[id_]->is_completed()){
      KALDI_WARN << "The first frame received, but it has requested terminate.";
      input_finished_ = true;
      break;
      // If you need safe coding, you can use following.
      //std::int32_t available_frames = socket_->buffers[id_]->get_available_frames();
      //if(available_frames == 0){
      //  KALDI_WARN << "The first frame received, but it has requested terminate.";
      //  input_finished_ = true;
      //}
      //else{
      //  // If you reached else's route, it is a bug.
      //  // The case of reaching else's route is only when another thread tries to process data of the same client.
      //}
      //break;
    }
    multiplatform::time::ms_sleep(10); // It wait only when a retry occurs.
    _warning_print("srcdata is not ready");
  } while(true);

  {
    feat_bytes_ = socket_->buffers[id_]->get_feature_bytes();
    feat_dim_ = socket_->buffers[id_]->get_feature_dimension();
    //KALDI_LOG << "dim = " << feat_dim_;
    //CAUTION// Please be aware that if you set to true the input_finish_ flag here, the features will not be set.
    //CAUTION// if(socket_->buffers[id_]->is_completed()){ }
  }

  if(feature_info_->feature_type.empty()){
    frame_shift_in_seconds_ = 0.01f;
  }
  else if(feature_info_->feature_type == "mfcc"){
    frame_shift_in_seconds_ = feature_info_->mfcc_opts.frame_opts.frame_shift_ms / 1000.0f;
  }
  else if(feature_info_->feature_type == "plp"){
    frame_shift_in_seconds_ = feature_info_->plp_opts.frame_opts.frame_shift_ms / 1000.0f;
  }
  else if(feature_info_->feature_type == "fbank"){
    frame_shift_in_seconds_ = feature_info_->fbank_opts.frame_opts.frame_shift_ms / 1000.0f;
  }
  else{
    frame_shift_in_seconds_ = 0.01f; // unknown feature type uses default value.
 }

}

OnlineFeatureMfcnet::~OnlineFeatureMfcnet() {
  //socket_.close();
}

void OnlineFeatureMfcnet::Initialize() {
}

kaldi::int32 OnlineFeatureMfcnet::Dim() const
{
  return feat_dim_;
}

kaldi::int32 OnlineFeatureMfcnet::NumFramesReady() const
{
  return num_frames_;
}

bool OnlineFeatureMfcnet::IsLastFrame(kaldi::int32 frame) const
{
  ////if (input_finished_) { KALDI_WARN << "\nframe=" << frame << ", num_frames_=" << num_frames_ << ", features_.NumRow()=" << features_.NumRows(); }
  return (input_finished_ ? ((frame + 1) == num_frames_) : false);
}

void OnlineFeatureMfcnet::GetFrame(kaldi::int32 frame, kaldi::VectorBase<kaldi::BaseFloat> *feat)
{
  Receive();
  KALDI_ASSERT(frame >= 0 && frame < num_frames_);
  KALDI_ASSERT(feat->Dim() == Dim());
  feat->CopyFromVec(features_.Row(frame));
}

//#if (KALDI_DOUBLEPRECISION != 0) is not supported. if you use double, received socket data and goto copy feat matrix
void OnlineFeatureMfcnet::Receive()
{ // feat_dim_ == Dim()
  if(socket_ == nullptr){ return; }
  if(input_finished_ == true){ return; }

  std::vector<float> feat_buf = socket_->buffers[id_]->get_feature();
  //std::vector<float> mask_buf = socket_->buffers[id_]->get_mask(); // This is dummy function.
  std::int32_t available_frames = socket_->buffers[id_]->get_available_frames();

  if(available_frames <= num_frames_){
    // Since there are no frames yet to be acquired at this point, we can immediately exit the method.
    if(!socket_->buffers[id_]->is_completed()){
      // Confirm that reception of the feature is not finished.
      if(socket_->find_terminated_key(id_)){
        // Since the socket was disconnected without receiving the termination packet of the feature,
        // an warning message is output and a completion flag is set.
        //KALDI_WARN << "terminated frame was not received, but socket are disconnected.";
        input_finished_ = true;
      }
    }
    return;
  }

  // You can trust the results obtained with get_available_frames() unless get_feature() is executed.
  // If you know that multiple threads do not work for the same client session (same utterance),
  // the following checks are no need.
  //if(feat_buf.size() < (available_frames * feat_dim_)){
  //  available_frames = (feat_buf.size() / feat_dim_);
  //}
  if(available_frames > 0){
    kaldi::Matrix<kaldi::BaseFloat> feats;
    //kaldi::Matrix<kaldi::BaseFloat> masks;

#if 1 // This is a method to preparing a new buffer and then swapping it with the old buffer.
    // This is initialization of Matrix, so kCopyData is not needed as the 3rd argument of Resize method.
    feats.Resize(available_frames, feat_dim_);
    //masks.Resize(available_frames, feat_dim_);

    for(int i=0; i<available_frames; i++){
      std::memcpy(feats.RowData(i), &feat_buf.at(feat_dim_*i), feat_bytes_);
      //std::memcpy(masks.RowData(i), mask_buf.at(feat_dim_*i), feat_bytes_);
    }

    if(available_frames != feats.NumRows()){
      KALDI_WARN << "available_frames=" << available_frames << ", feats.NumRows()=" << feats.NumRows();
    }

    KALDI_VLOG(3) << "source_id = " << source_info_.source_id
                  << ", received frames = " << available_frames;

    features_.Swap(&feats);
    num_frames_ = available_frames;
#else // This is a buffer resizing and joining method similar to conventional behavior.
    std::int32_t increased_frames = available_frames - num_frames_;
    // This is initialization of Matrix, so kCopyData is not needed as the 3rd argument of Resize method.
    feats.Resize(increased_frames, feat_dim_);
    //masks.Resize(increased_frames, feat_dim_);

    for(int i=0, j=num_frames_; i<increased_frames; i++, j++){
      std::memcpy(feats.RowData(i), &feat_buf.at(feat_dim_*j), feat_bytes_);
      //std::memcpy(masks.RowData(i), mask_buf.at(feat_dim_*j), feat_bytes_);
    }

    if(increased_frames != feats.NumRows()){
      KALDI_WARN << "increased_frames=" << increased_frames << ", feats.NumRows()=" << feats.NumRows();
    }

    KALDI_VLOG(3) << "source_id = " << source_info_.source_id
                  << ", received frames = " << available_frames;

    std::int32_t new_num_frames = num_frames_ + feats.NumRows(); // feats.NumRows() is increased_frames.
    const float increase_ratio = 1.5f;
    if(new_num_frames > features_.NumRows()){
      kaldi::int32 new_num_rows = std::max<kaldi::int32>(new_num_frames, features_.NumRows() * increase_ratio);
      features_.Resize(new_num_rows, feat_dim_, kCopyData);
    }
    features_.Range(num_frames_, feats.NumRows(), 0, feat_dim_).CopyFromMat(feats);
    num_frames_ = new_num_frames;
#endif // SELECT_***_MODE

  }
  if(socket_->buffers[id_]->is_completed()){ input_finished_ = true; }
}

} // namespace kaldi
