// recognizer.h

#ifndef KALDI_RECOGNIZER1_H_
#define KALDI_RECOGNIZER1_H_

#include <memory>

#include "decoder/lattice-faster-online-decoder.h"
#include "nnet/nnet-nnet.h"
#include "nnet/nnet-pdf-prior.h"
#include "decodable-nnet1-online-for-hark.h"
#include "lat/word-align-lattice-lexicon.h"
#include "lat/lattice-functions.h"

#include "recognizer0.h"

#include <iomanip>
#include <string>
#include <vector>

namespace kaldi {
namespace nnet1 {

class NnetRecognizer : kaldi::Recognizer {
 public:
  NnetRecognizer(const kaldi::nnet1::Nnet& nnet,
                 kaldi::nnet1::PdfPrior& pdf_prior,
                 const kaldi::TransitionModel& trans_model,
                 kaldi::nnet1::DecodableNnetOnlineForHarkOptions opt,
                 std::shared_ptr<kaldi::OnlineFeatureMfcnet>& features,
                 std::shared_ptr<kaldi::OnlineFeatureInterface>& feature_if,
                 const fst::Fst<fst::StdArc>* decode_fst,
                 const kaldi::LatticeFasterDecoderConfig& decoder_config,
                 const fst::SymbolTable* phone_syms,
                 const fst::SymbolTable* word_syms,
                 const kaldi::WordAlignLatticeLexiconInfo& lexicon_info,
                 std::string& lm_name,
                 int emulate_julius,
                 int nbest_count,
                 int progress_output_interval,
                 std::shared_ptr<result_server> result_client = nullptr);
  virtual ~NnetRecognizer() {}

  virtual std::shared_ptr<OnlineFeatureInterface> GetOnlineFeatureInterface();
  virtual void Recognize();
  virtual void WriteIntermediateResult(kaldi::int32);
  virtual void WriteResult();
  virtual void PrintResult(std::ostream& out);

  void SetLmName(const std::string& lm_name);

 private:
  const kaldi::TransitionModel& trans_model_;
  std::shared_ptr<kaldi::OnlineFeatureMfcnet> features_;
  std::shared_ptr<kaldi::OnlineFeatureInterface> features_if_;
  kaldi::nnet1::DecodableNnetOnlineForHark decodable_;
  kaldi::LatticeFasterOnlineDecoder decoder_;
  const fst::SymbolTable* phone_syms_;
  const fst::SymbolTable* word_syms_;
  const kaldi::WordAlignLatticeLexiconInfo& lexicon_info_;
  std::shared_ptr<result_server> result_client_;
  std::string lm_name_;
  int emulate_julius_;
  int nbest_count_;
  int progress_output_interval_;
  std::string delimiter_sym_;
  bool features_input_is_finished;
};

template<class Arc, class I>
bool GetPhoneAlignmentSequence(const fst::Fst<Arc> &fst,
                               std::vector<I> *phones_out) {
  //using namespace fst;
  //typedef typename Arc::Label Label;
  //typedef typename Arc::StateId StateId;
  //typedef typename Arc::Weight Weight;

  std::vector<I> phones_seq;

  typename Arc::StateId cur_state = fst.Start();
  if(cur_state == fst::kNoStateId){  // empty sequence.
    if(phones_out != NULL){ phones_out->clear(); }
    return true;
  }
  while(1){
    typename Arc::Weight w = fst.Final(cur_state);
    if(w != Arc::Weight::Zero()){  // is final..
      if(fst.NumArcs(cur_state) != 0){ return false; }
      if(phones_out != NULL){ *phones_out = phones_seq; }
      return true;
    }
    else{
      if(fst.NumArcs(cur_state) != 1){ return false; }

      fst::ArcIterator<fst::Fst<Arc> > iter(fst, cur_state);  // get the only arc.
      const Arc &arc = iter.Value();
      if(arc.ilabel != 0){ phones_seq.push_back(arc.ilabel); }
      cur_state = arc.nextstate;
    }
  }
}

template<class I>
void GetWordAlignmentSequence(const kaldi::Lattice &best_path_lat,
                              const kaldi::TransitionModel& trans_model,
                              const kaldi::WordAlignLatticeLexiconInfo& lexicon_info,
                              std::vector<I> *words_out,
                              kaldi::LatticeWeight *weight) {
  std::vector<I> alignment;
  kaldi::CompactLattice best_path_clat;
  kaldi::CompactLattice best_path_clat_det;
  kaldi::CompactLattice aligned_clat;
  kaldi::CompactLattice best_path_aligned_clat;
  kaldi::Lattice best_path_aligned_lat;
  kaldi::WordAlignLatticeLexiconOpts lexicon_opts;
  
  // Lattice => CompactLattice => remove_eps = determinize = aligned = best_path => CompactLattice => Lattice
  fst::ConvertLattice(best_path_lat, &best_path_clat);
  fst::RemoveEpsLocal(&best_path_clat);
  fst::Determinize(best_path_clat, &best_path_clat_det);
  kaldi::WordAlignLatticeLexicon(best_path_clat_det, trans_model, lexicon_info, lexicon_opts, &aligned_clat);
  kaldi::CompactLatticeShortestPath(aligned_clat, &best_path_aligned_clat);
  fst::ConvertLattice(best_path_aligned_clat, &best_path_aligned_lat);
  fst::GetLinearSymbolSequence(best_path_aligned_lat, &alignment, words_out, weight);
}

template<class I, class J>
void GetWordWithPhoneAlignmentSequence(const kaldi::Lattice &best_path_lat,
                                       const kaldi::TransitionModel& trans_model,
                                       const kaldi::WordAlignLatticeLexiconInfo& lexicon_info,
                                       std::vector<I> *words_out,
                                       std::vector<J> *phones_out,
                                       kaldi::LatticeWeight *weight) {
  std::vector<I> alignment;
  kaldi::CompactLattice best_path_clat;
  kaldi::CompactLattice best_path_clat_det;
  kaldi::CompactLattice aligned_clat;
  kaldi::CompactLattice best_path_aligned_clat;
  kaldi::Lattice best_path_aligned_lat;
  kaldi::WordAlignLatticeLexiconOpts lexicon_opts;
  
  // Lattice => CompactLattice => remove_eps = determinize = aligned = best_path => CompactLattice => Lattice
  fst::ConvertLattice(best_path_lat, &best_path_clat);
  fst::RemoveEpsLocal(&best_path_clat);
  fst::Determinize(best_path_clat, &best_path_clat_det);
  kaldi::WordAlignLatticeLexicon(best_path_clat_det, trans_model, lexicon_info, lexicon_opts, &aligned_clat);
  kaldi::CompactLatticeShortestPath(aligned_clat, &best_path_aligned_clat);
  fst::ConvertLattice(best_path_aligned_clat, &best_path_aligned_lat);
  fst::GetLinearSymbolSequence(best_path_aligned_lat, &alignment, words_out, weight);
}
} // namespace nnet1
} // namespace kaldi

#endif // KALDI_RECOGNIZER1_H_
